/**
 * COMP 401 Fall 2017 -- Recitation 10 November 2017
 * 
 * Copyright (C) 2017, Alex Blate. All Rights Reserved.
 */
package recitations.recitation11;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Objects;

/**
 * FIFO queue implemented as a doubly-linked list
 * 
 */
public class LinkedQueue<T> extends AbstractQueue<T> {
	
	/**
	 * list.next <-> first in line <-> second in line <-> ... <-> last in line <-> list
	 * list.prev <-> last in line <-> ... <-> second in line <-> first in line <-> list
	 */
	protected ListNode<T> list = new ListNode<T>();
	protected int size=0;

	public LinkedQueue() {
		list.setNext(list);
		list.setPrev(list);
	}		

	/**
	 * State before push:
	 * 		list.prev <-> OLD_TAIL <-> ...
	 * State after push:
	 * 		list.prev <-> [e] <-> OLD_TAIL <->
	 *     
	 */
	@Override
	public void enqueue(T e) {
		ListNode<T> oldTail = getTail();
		
		ListNode<T> elt = new ListNode<>(e, getTail(), list);
		getTail().setNext(elt);
		setTail(elt);
		
		assert (oldTail.getNext()==elt);
		assert (elt.getPrev()==oldTail);
		assert (elt.getNext() == list);
		assert (getTail() == elt);
		assert (list.validLinks() && elt.validLinks() && oldTail.validLinks());
		
		size++;
		
	}


	/**
	 * Remove and return the element at the head of the queue or
	 * return null if the queue is empty.
	 * 
	 */
	@Override
	public T dequeue() {
		T res = null;
		if(size > 0 ) {
			
			ListNode<T> oldHead = getHead();
			assert list != oldHead;
			
			res = oldHead.getValue();
			oldHead.getNext().setPrev(list);
			list.setNext(oldHead.getNext());
			
			oldHead.setValue(null);
			oldHead.setNext(null);
			oldHead.setPrev(null);
			
			size--;
			
			// Is this check equivalent to (size != 0 ) && list.validLinks() ? 
			assert (size != 0) || ( list.getNext() == list && list.getPrev()==list);			
		}
		
		assert list.validLinks();
		return res;
	}

	@Override
	public T head() {
		return getHead().getValue();
	}

	@Override
	public int size() {
		return size;
	}
	
	public Object[] toArray() {
		Object[] res = new Object[size()];

		ListNode<T> it = list.getNext();
		int i = 0;
		
		//TODO: Maybe should implement a proper iterator
		while(it != list) {			
			res[i] = it.getValue();
			it = it.getNext();
			i++;
		}
		
		return res;
	}
	
	// What if I wanted a method that returned T[] instead of Object[]
	// Something along the lines of:
	//       public T[] toArray() 
	// How do I do this??
	//
	// The way I know...
	
	public T[] toArray(T[] dummy) {
		Objects.requireNonNull(dummy);
		
		@SuppressWarnings("unchecked")
		T[] res = (T[]) Array.newInstance(dummy.getClass(), size());
		ListNode<T> it = list.getNext();
		int i = 0;
		while(it != list) {			
			res[i] = it.getValue();
			it = it.getNext();
			i++;
		}
		
		return res;
	}
	
	
	
	@Override
	public String toString() {
		StringBuffer res = new StringBuffer();
		res.append("Queue: size=").append(size()).append(" ");
		if(!isEmpty() ) {
			res.append("head=<");
			if(getHead().getValue() == null) {
				res.append("null");
			} else {
				res.append(getHead().getValue().toString());
			}
			res.append("> Contents: ");
		}
		
		ArrayList<String> values = new ArrayList<>(size());
		ListNode<T> it = list.getNext();
		while(it != list) {
			if(it.getValue() == null) {
				values.add("null");
			} else  {				
				values.add(it.getValue().toString());
			}
			
			it = it.getNext();
		}
		
		res.append(Arrays.toString(values.toArray()));
		return res.toString();
	}//toString()

	
	protected ListNode<T> getHead() {
		return list.getNext();
	}
	
	protected ListNode<T> getTail() {
		return list.getPrev();
	}
	
	protected void setTail(ListNode<T> tail) {
		list.setPrev(tail);
	}
	
	protected void setHead(ListNode<T> head) {
		list.setNext(head);
	}
	

	
}//class
